import argparse

from . import api
from . import utils

if __name__ == "__main__":                                                      
    parser = argparse.ArgumentParser(description = 'utils')                     
    parser.add_argument("-w", "--what", help="command to run")
    parser.add_argument("-t", "--tenant", help="tenant")
    parser.add_argument("-v", "--verbose", action="store_true", help="verbose")
    parser.add_argument("-o", "--op", choices=["DEL", "ADD"], help="verbose")
    parser.add_argument("-q", "--quiet", action="store_true", help="count by unique values")
    parser.add_argument("-c", "--countby", help="count by unique values")       
    parser.add_argument("-g", "--groupby", help="groub by (key)")               
    parser.add_argument("-f", "--filterby", help="filter by (key)")             
    args = parser.parse_args() 

    if args.verbose: api.g_debug = True
    api = api.API()

    if args.what == "deltenant":
        print(api.delete_tenant(tenant=args.tenant))
    if args.what == "enforced":
        assert args.tenant, "need to specify tenant"
        if input("are you sure?> ") == "yes":
            assert args.op, "need to specify op"
            api.set_enforced(op=args.op, zone="poczone", app="kube-system", pod="araali-fw", container="araali-fw",
                tenant=args.tenant)

        utils.dump_table(api.get_enforced(args.tenant)[0], quiet=args.quiet,
                filterby=args.filterby, countby=args.countby, groupby=args.groupby)



from __future__ import print_function

import os
import shlex
import subprocess
import sys
import yaml

cfg_fname = os.environ['HOME']+"/.araalirc"

def update_kubeconfig(cluster_name):
    return run_command_logfailure("aws eks update-kubeconfig --name %s" % cluster_name)

def flatten_obj(prefix, root):
    res = []
    for k, v in root.items():
        new_prefix = k if not prefix else prefix + "." + k
        if type(v) == dict:
            res += flatten_obj(new_prefix, v)
        else:
            res += [(new_prefix, str(v))]
    return res

def flatten_table(alerts):
    for a in alerts:
        yield(dict(flatten_obj("", a)))

def dump_csv(alerts):
    header = set()
    for a in alerts:
        header |= set(dict(flatten_obj("", a)).keys())
    header = list(header)
    header.sort()
    yield ",".join(header)
    for a in alerts:
        obj = dict(flatten_obj("", a))
        out = []
        for h in header:
            out.append(obj.get(h, ""))
        yield ",".join(out)

def read_config():
    if os.path.isfile(cfg_fname):
        with open(cfg_fname, "r") as f:
            cfg = yaml.load(f, Loader=yaml.SafeLoader)
    else:
        cfg = {}

    if not cfg.get("tenant", None): cfg["tenant"] = None
    if not cfg.get("template_dir", None): cfg["template_dir"] = "../templates"
    if not cfg.get("token", None): cfg["token"] = os.getenv('ARAALI_API_TOKEN')
    if not cfg.get("backend", None): cfg["backend"] = "prod"
    # env variable overrides on disk config
    if os.getenv("ARAALI_BACKEND"):
        cfg["backend"] = os.getenv("ARAALI_BACKEND")
    return cfg

cfg = read_config()

def config(tenant=None, tenants=None, template_dir=None, backend=None):
    global cfg

    def save(cfg):
        with open(cfg_fname, "w") as f:
            cfg2 = dict(cfg)
            del cfg2["token"]
            yaml.dump(cfg2, f)

    if tenant is not None:
        if not tenant: # passed as empty string
            tenant = None # store it as nil in yaml
        cfg["tenant"] = tenant
        save(cfg)
    if tenants is not None:
        cfg["tenants"] = [dict(zip(["name", "id"], a.split(":"))) for a in tenants.split(",")]
        save(cfg)
    for k in ["template_dir", "backend"]:
        if eval(k) is not None:
            cfg[k] = eval(k)
            save(cfg)
    print(yaml.dump(cfg))

def get_by_key(o, k):
    k = k.split(".")
    for a in k:
        o = o.get(a, None)
        if o is None: return "None"
    return o if o is not None else "None"

def dump_table(objs, quiet=False, filterby=None, countby=None, groupby=None, dump_yaml=False, flatten=False):
    class FilterBy(object):
        def __init__(self, filterby):
            if filterby:
                self.filterby = [a.split("=") for a in filterby.split(":")]
            else:
                self.filterby = None

        def filter(self, o):
            if self.filterby is None: return True

            match_count = 0
            for f in self.filterby:
                k, v = f
                for a in v.split(","): # like an OR
                    def get_a(a):
                        try:
                            return eval(a)
                        except:
                            return a
                    match_val = get_a(a)
                    my_val = get_by_key(o, k)
                    if type(match_val) == str and ".*" in match_val and re.search(match_val, my_val) or my_val == match_val:
                        match_count += 1
                        break
            # matched all filters, like AND
            if match_count == len(self.filterby):
                return True
            return False

    class CountBy(object):
        def __init__(self, countby, groupby):
            self.countby = countby.split(",") if countby else countby
            self.groupby = groupby.split(",") if groupby else groupby
            if countby:
                if groupby is None:
                    self.groupby_dict = {"global": {}}
                    for c in self.countby:
                        self.groupby_dict["global"][c] = 0
                else:
                    self.groupby_dict = {}

        def count(self, o):
            if self.countby:
                if self.groupby:
                    groupby_key = []
                    for g in self.groupby:
                        groupby_key.append(str(get_by_key(o, g)))
                    groupby_key = ":".join(groupby_key)
                    if groupby_key not in self.groupby_dict:
                        self.groupby_dict[groupby_key] = {}
                        for c in self.countby:
                            self.groupby_dict[groupby_key][c] = 0
                else:
                    groupby_key = "global"

                for k,v in self.groupby_dict[groupby_key].items():
                    value = get_by_key(o, k)
                    if type(value) == int:
                        self.groupby_dict[groupby_key][k] += value
                    else:
                        self.groupby_dict[groupby_key][k] += 1

        def show(self, dump_yaml=False):
            if not self.countby:
                return

            if dump_yaml:
                print(yaml.dump([{"key": a[0], "count": a[1]} for a in self.groupby_dict.items()]))
                return

            overall = {}
            for g, countby_dict in self.groupby_dict.items():
                print("\n")
                print("="*40, "%s" % g, "="*40)
                if countby:
                    for k,v in countby_dict.items():
                        print("%ss: %s" % (k, v))
                        overall[k] = overall.get(k, 0) + v

            print("="*40, "%s" % "overall", "="*40)
            for k,v in overall.items():
                print("%ss: %s" % (k, v))

    countby = CountBy(countby, groupby)
    filterby = FilterBy(filterby)
    for idx, o in enumerate(objs):
        if not filterby.filter(o): continue
        if not quiet:
            print("%s %s %s" % ("="*40, idx, "="*40))
            if flatten:
                print(yaml.dump(dict(flatten_obj("", o))))
            else:
                print(yaml.dump(o))
        countby.count(o)
    countby.show(dump_yaml)

def make_map(kvstr):
    k, v = kvstr.split("=")
    return k, int(v)

def eprint(*args, **kwargs):
    args = [a.decode() for a in args]
    print(*args, file=sys.stderr, **kwargs)

def run_command_logfailure(cmd):
    rc = run_command(cmd, debug=False, result=True, strip=False)
    if rc[0] != 0:
        print("*** failed: %s" % rc[1].decode())
        return False, None
    return True, rc[1].decode()

def run_command(command, result=False, strip=True, in_text=None, debug=False, env=os.environ):
    def collect_output(ret, output):
        if output:
            if strip:
                output = output.strip()
            if result:
                ret.append(output.decode())
            else:
                eprint(output)

    if debug:
        print(command, result)
    process = subprocess.Popen(shlex.split(command), stdout=subprocess.PIPE,
                                stdin=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                env=env)
    ret = []
    if result:
        if in_text:
            outs, errs = process.communicate(input=in_text.encode())
        else:
            outs, errs = process.communicate()
        return process.poll(), outs+errs
    else:
        while True:
            rc = process.poll()
            err = process.stderr.readline()
            out = process.stdout.readline()
            if rc is not None and not out and not err:
                return rc, "\n".join(ret)
            collect_output(ret, out)
            collect_output(ret, err)



#!/usr/bin/env python

mystr = """
/araali_fog2agent.araalifog2agentservices/attesttargetworkload:
  Count: 25755
  Errors: 0
  Avg: 11269
  Min: 0
  Max: 1646546
  InBytes: 75147845
  OutBytes: 312561961
/araali_fog2agent.araalifog2agentservices/reportprocessvulnerabilities:
  Count: 1075
  Errors: 0
  Avg: 57169
  Min: 0
  Max: 1259702
  InBytes: 6742865
  OutBytes: 7525
/araali_fog2agent.araalifog2agentservices/imagedownload:
  Count: 11
  Errors: 0
  Avg: 17
  Min: 0
  Max: 25
  InBytes: 9761
  OutBytes: 773090532
/araali_fog2agent.araalifog2agentservices/reportassets:
  Count: 121
  Errors: 0
  Avg: 4
  Min: 0
  Max: 45
  InBytes: 152873
  OutBytes: 847
/araali_fortify.araalifogfortifyservices/listk8sservicesbyoperator:
  Count: 2
  Errors: 0
  Avg: 173743
  Min: 0
  Max: 264459
  InBytes: 136
  OutBytes: 18
/araali_fortify.araalifogfortifyservices/registerworkload:
  Count: 1
  Errors: 0
  Avg: 17820
  Min: 0
  Max: 17820
  InBytes: 259
  OutBytes: 54
/araali_fog2agent.araalifog2agentservices/getsyscallrules:
  Count: 3
  Errors: 0
  Avg: 5
  Min: 0
  Max: 6
  InBytes: 210
  OutBytes: 339
/araali_fog2agent.araalifog2agentinitialservices/attestworkload:
  Count: 9
  Errors: 0
  Avg: 2716
  Min: 0
  Max: 2793
  InBytes: 22198
  OutBytes: 104932
/araali_fog2agent.araalifog2agentservices/sendheartbeat:
  Count: 267788
  Errors: 0
  Avg: 4
  Min: 0
  Max: 14138
  InBytes: 34322584
  OutBytes: 4016820
/araali_fog2agent.araalifog2agentservices/getaraalihostsv2:
  Count: 263290
  Errors: 0
  Avg: 375
  Min: 0
  Max: 14246
  InBytes: 127033099
  OutBytes: 44881246877
/araali_fog2agent.araalifog2agentservices/sendgtorinfo:
  Count: 65824
  Errors: 0
  Avg: 23
  Min: 0
  Max: 12349
  InBytes: 175292402
  OutBytes: 1261991
/araali_fog2agent.araalifog2agentservices/getaraalipolicy:
  Count: 40796
  Errors: 0
  Avg: 143
  Min: 0
  Max: 48416
  InBytes: 113958861
  OutBytes: 169804848
/araali_fog2agent.araalifog2agentservices/reportflows:
  Count: 214983
  Errors: 0
  Avg: 5
  Min: 0
  Max: 5354
  InBytes: 5568400893
  OutBytes: 2149820
/araali_fog2agent.araalifog2agentservices/servicesession:
  Count: 38
  Errors: 0
  Avg: 9
  Min: 0
  Max: 46
  InBytes: 64309
  OutBytes: 1464
/araali_fortify.araalifogfortifyservices/registerservice:
  Count: 74
  Errors: 0
  Avg: 7527
  Min: 0
  Max: 142585
  InBytes: 112864
  OutBytes: 518
Total:
  InBytes: 6101261159
  OutBytes: 46144248546
Flow Count: 1892168 (2022-08-25 02:44:33.495749646 +0000 UTC m=+47206.162003645)
CPU Usage: 2.88%/2.83%
Memory Usage: 4.85%/4.85%
A2B Queue Stats:
   QElemTypeAgentInfo: 0
   QElemTypeAgentHeartbeat: 0
   QElemTypeFlowInfo: 19
   QElemTypeFortificationInfo: 0
   QElemTypeTimer: 0
   QElemTypeMalwareAlert: 19
   QElemTypeFetchPolicy: 0
   QElemTypeAgentHeartbeatV2: 0
   QElemTypeListK8Services: 0
   QElemTypeAssetInfo: 0
   QElemTypeProcessInfo: 0
   QElemTypeAraaliK8SPolicy: 0
   invalid q type: 0
   """
hdr = ["Count", "Errors", "Avg", "Min", "Max", "InBytes", "OutBytes"]
topic = None
items = {}
for line in mystr.strip().split("\n"):
    line = line.strip().split(":")
    if not line[1]:
        topic = line[0]
        items[topic] = {}
        continue
    if line[0] in hdr:
        items[topic][line[0]] = line[1]
        continue
    print(line)

print(",".join(hdr))
for k,v in items.items():
    print(k, end=",", sep='')
    for h in hdr:
        print(v.get(h, 0), end=",", sep='')
    print()



import boto3
import email as _email
import os
import sys

from . import utils

def cf_ls(get_all):
        client = boto3.client("cloudformation")
        ret = client.list_stacks()
        while "NextToken" in ret:
            for stack in ret["StackSummaries"]:
                if get_all or "araaliapicf-" in stack["StackName"]:
                    if stack["StackStatus"] in ["DELETE_COMPLETE"]: continue
                    obj = {"id": stack["StackId"], "name": stack["StackName"],
                            "Created": stack["CreationTime"],
                            "Status": stack["StackStatus"],
                            "Drift": stack["DriftInformation"],
                            }
                    outputs = client.describe_stacks(StackName=stack["StackName"])["Stacks"][0]
                    if "Outputs" in outputs:
                        obj["Outputs"] = outputs["Outputs"]
                    yield obj
            ret = client.list_stacks(NextToken=ret["NextToken"])

def cf_rm(name):
    print("removing %s" % name)
    client = boto3.client("cloudformation")
    ret = client.delete_stack(StackName=name)
    print(ret)

def cf_add_vm():
    print("adding vm")

def verify_email(email):
    if email is None: return False
    email = _email.utils.parseaddr(email)[1].split("@")
    return len(email) >= 2 and email[0] and email[1]

def verify_key_pair(key_pair):
    if key_pair is None: return False
    key_pair = os.path.basename(key_pair)
    key_pair = key_pair.rsplit(".", 1)[0]

    ec2 = boto3.client('ec2')
    ret = ec2.describe_key_pairs()
    valid_keys = [a["KeyName"] for a in ret["KeyPairs"]]
    if key_pair not in valid_keys:
        for k in valid_keys:
            if key_pair in k:
                print("*** did you mean: %s" % k)
                sys.exit(1)
        print("*** valid keys must be one of: \n\t%s" % "\n\t".join(valid_keys))
        sys.exit(1)
    return key_pair, key_pair in [a["KeyName"] for a in ret["KeyPairs"]]

def cf_launch_fortified_vm(stack_name, email, key_pair, ami_id):
    if not verify_email(email):
        print("*** email: %s not valid" % email)
        sys.exit(1)

    key_pair, kp_valid = verify_key_pair(key_pair)
    if not kp_valid:
        print("*** key-pair: %s not valid" % key_pair)
        sys.exit(1)

    if ami_id is None:
        ssm_client = boto3.client('ssm')
        response = ssm_client.get_parameter(Name='/aws/service/ami-amazon-linux-latest/amzn2-ami-hvm-x86_64-gp2')
        if "Parameter" not in response and "Value" not in response["Parameter"]:
            print("Error: Failed to fetch latest AmazonLinux2 AMI ID from SSM")
            return False, None
        ami_id = response["Parameter"]["Value"]

    template_url = "https://s3.us-west-1.amazonaws.com/araalinetworks.test/quickstart_vm/AraaliVMQuickstartInternalStack.template.json"
    return utils.run_command_logfailure(
        "aws cloudformation create-stack --stack-name %s --template-url %s\
        --parameters ParameterKey=KeyName,ParameterValue=%s ParameterKey=email,ParameterValue=%s ParameterKey=AmiId,ParameterValue=%s" % (
        stack_name, template_url, key_pair, email, ami_id)
    )

def cf_launch_eks_cluster(stack_name, cluster_name, availability_zones):
    template_url = "https://s3.amazonaws.com/aws-quickstart/quickstart-amazon-eks/templates/amazon-eks-entrypoint-new-vpc.template.yaml"
    availability_zone_str = "\\\\,".join(availability_zones)
    return utils.run_command_logfailure(
       "aws cloudformation create-stack --stack-name %s --template-url %s --parameters ParameterKey=AvailabilityZones,ParameterValue=%s\
     ParameterKey=EKSClusterName,ParameterValue=%s ParameterKey=NumberOfAZs,ParameterValue=%s ParameterKey=EKSPublicAccessEndpoint,ParameterValue=Enabled --capabilities CAPABILITY_IAM CAPABILITY_AUTO_EXPAND" % (
        stack_name, template_url, availability_zone_str, cluster_name, len(availability_zones))
    )

def cf_validate_stack_creation(stack_name):
    cf_client = boto3.client('cloudformation')
    cf_response = cf_client.describe_stacks(StackName=stack_name)
    cf_stacks = cf_response["Stacks"]
    cf_stack = cf_stacks[0]
    stack_status = cf_stack["StackStatus"]
    return stack_status ==  "CREATE_COMPLETE"

def assets(members):
    curr_account = boto3.client('sts').get_caller_identity().get('Account')
    if members:
        def get_accounts():
            for acc in boto3.client('organizations').list_accounts()["Accounts"]:
                yield acc
        accounts = get_accounts()
    else:
        accounts = [{"Id": curr_account}]
    for acc in accounts:
        if acc["Id"] == curr_account:
            session = boto3.session.Session()
        else:
            role_info = {
                'RoleArn': 'arn:aws:iam::%s:role/OrganizationAccountAccessRole' % acc["Id"],
                'RoleSessionName': 'araaliapi'
            }
            credentials = boto3.client('sts').assume_role(**role_info)
            session = boto3.session.Session(
                aws_access_key_id=credentials['Credentials']['AccessKeyId'],
                aws_secret_access_key=credentials['Credentials']['SecretAccessKey'],
                aws_session_token=credentials['Credentials']['SessionToken']
            )

        ec2_client = session.resource("ec2")
        for instance in ec2_client.instances.all():
            if 0:
                for attr in dir(instance):
                    if attr[0] != "_":
                        print("{field: %s, value: %s}" % (attr, getattr(instance, attr)))
                return
            # 'ami_launch_index', 'architecture', 'attach_classic_link_vpc',
            # 'attach_volume', 'block_device_mappings', 'boot_mode',
            # 'capacity_reservation_id', 'capacity_reservation_specification',
            # 'classic_address', 'client_token', 'console_output',
            # 'cpu_options', 'create_image', 'create_tags', 'delete_tags',
            # 'describe_attribute', 'detach_classic_link_vpc', 'detach_volume',
            # 'ebs_optimized', 'elastic_gpu_associations',
            # 'elastic_inference_accelerator_associations', 'ena_support',
            # 'enclave_options', 'get_available_subresources',
            # 'hibernation_options', 'hypervisor', 'iam_instance_profile',
            # 'id', 'image', 'image_id', 'instance_id', 'instance_lifecycle',
            # 'instance_type', 'ipv6_address', 'kernel_id', 'key_name',
            # 'key_pair', 'launch_time', 'licenses', 'load',
            # 'maintenance_options', 'meta', 'metadata_options',
            # 'modify_attribute', 'monitor', 'monitoring',
            # 'network_interfaces', 'network_interfaces_attribute',
            # 'outpost_arn', 'password_data', 'placement', 'placement_group',
            # 'platform', 'platform_details', 'private_dns_name',
            # 'private_dns_name_options', 'private_ip_address',
            # 'product_codes', 'public_dns_name', 'public_ip_address',
            # 'ramdisk_id', 'reboot', 'reload', 'report_status',
            # 'reset_attribute', 'reset_kernel', 'reset_ramdisk',
            # 'reset_source_dest_check', 'root_device_name',
            # 'root_device_type', 'security_groups', 'source_dest_check',
            # 'spot_instance_request_id', 'sriov_net_support', 'start',
            # 'state', 'state_reason', 'state_transition_reason', 'stop',
            # 'subnet', 'subnet_id', 'tags', 'terminate', 'tpm_support',
            # 'unmonitor', 'usage_operation', 'usage_operation_update_time',
            # 'virtualization_type', 'volumes', 'vpc', 'vpc_addresses',
            # 'vpc_id', 'wait_until_exists', 'wait_until_running',
            # 'wait_until_stopped', 'wait_until_terminated'
            #print(dir(instance))
            if instance.state["Name"] == "stopped": continue
            yield {
                    "account": acc["Id"],
                    "image_id": instance.image.id,
                    "instance_id": instance.id,
                    "key_name": instance.key_name,
                    "launch-time": instance.launch_time,
                    "state": instance.state["Name"],
                    "subnet": instance.subnet_id,
                    "platform": instance.platform,
                    "public_ip": instance.public_ip_address,
                    "security_groups": instance.security_groups,
                    "tags": instance.tags,
                    "type": instance.instance_type,
                    "vpc": instance.vpc_id,
            }



from __future__ import print_function

import datetime
import requests
import json
import yaml

from . import utils

g_debug = False

class API:
    def __init__(self):
        if utils.cfg["token"] is None:
            raise Exception("*** ARAALI_API_TOKEN environment variable is not set")

        self.host = "https://api-%s.aws.araalinetworks.com" % utils.cfg["backend"]
        self.headers = {"Authorization": "Bearer %s" % utils.cfg["token"]}

    def get(self, endpoint, data):
        if g_debug: print(self.host, self.headers, endpoint)
        rc = requests.get("%s/%s" % (self.host, endpoint), params=data, headers=self.headers)
        if g_debug: print(rc.request.url)
        if rc.status_code != 200:
            print("*** REST api error: %s" % rc.status_code)
        return rc.json(), 0 if rc.status_code == 200 else rc.status_code

    def post(self, endpoint, data):
        if g_debug: print(self.host, self.headers, endpoint)
        rc = requests.post("%s/%s" % (self.host, endpoint),
                           json=data, headers=self.headers)
        if g_debug:
            print(rc.request.url)
            print(rc.json())
        if rc.status_code != 200:
            print("*** REST api error: %s" % rc.status_code)
        return rc.json(), 0 if rc.status_code == 200 else rc.status_code

    def get_alerts(self, count=None, ago=None, token=None, closed_alerts=False, all_alerts=False, tenant=None):
        """Fetches alerts
            Usage: alerts, next_page, status = api.get_alerts()
        """
        data = {
                "filter.openAlerts": True,
                "filter.closedAlerts": closed_alerts,
                "filter.listAllAlerts": all_alerts,
                "filter.perimeterIngress": True,
                "filter.perimeterEgress": True,
                "filter.homeNonAraaliIngress": True,
                "filter.homeNonAraaliEgress": True,
                "filter.araaliToAraali": True,
        }
        if tenant is None:
            if utils.cfg["tenant"]: data["tenant.id"] = utils.cfg["tenant"]
        else:
            data["tenant.id"] = tenant

        if not ago:
            ago = "infinite"
        if ago != "infinite":
            ago = dict([utils.make_map(a) for a in ago.split(",")])
            data["filter.time.startTime"] = (datetime.datetime.utcnow() - datetime.timedelta(**ago)).strftime('%Y-%m-%dT%H:%M:%SZ')
            data["filter.time.endTime"] = (datetime.datetime.utcnow()).strftime('%Y-%m-%dT%H:%M:%SZ')

        if count is None: count = 1000
        data["count"] = count

        if token: data["pagingToken"] = token
        if g_debug: print(data)

        ret, status = self.get("api/v2/listAlerts", data)
        return ret.get("links", []), ret.get("paging_token", None), status

    def get_assets(self, zone=None, app=None, ago=None, tenant=None):
        """Fetches assets
            Usage: assets, status = api.get_assets()
        """
        data = {
                "filter.listActiveVm": True,
                "filter.listInactiveVm": False,
                "filter.listActiveContainer": True,
                "filter.listInactiveContainer": False,
        }
        if tenant is None:
            if utils.cfg["tenant"]: data["tenant.id"] = utils.cfg["tenant"]
        else:
            data["tenant.id"] = tenant
        
        if zone: data["zone"] = zone
        if app: data["app"] = app

        if not ago:
            ago = "days=1"
        ago = dict([utils.make_map(a) for a in ago.split(",")])
        data["time.startTime"] = (datetime.datetime.utcnow() - datetime.timedelta(**ago)).strftime('%Y-%m-%dT%H:%M:%SZ')
        data["time.endTime"] = (datetime.datetime.utcnow()).strftime('%Y-%m-%dT%H:%M:%SZ')
        
        if g_debug: print(data)

        ret, status = self.get("api/v2/listAssets", data)
        return ret.get("assets", []), status

    def get_links(self, zone=None, app=None, svc=None, ago=None, tenant=None):
        """Fetches links
            Usage: links, status = api.get_links()
        """
        data = {}
        if tenant is None:
            if utils.cfg["tenant"]: data["tenant.id"] = utils.cfg["tenant"]
        else:
            data["tenant.id"] = tenant
        if zone: data["zone"] = zone
        if app: data["app"] = app
        if svc: data["service"] = svc

        if not ago:
            ago = "days=1"
        ago = dict([utils.make_map(a) for a in ago.split(",")])
        data["time.startTime"] = (datetime.datetime.utcnow() - datetime.timedelta(**ago)).strftime('%Y-%m-%dT%H:%M:%SZ')
        data["time.endTime"] = (datetime.datetime.utcnow()).strftime('%Y-%m-%dT%H:%M:%SZ')
        
        if g_debug: print(data)

        ret, status = self.get("api/v2/listLinks", data)
        return ret.get("links", []), status

    def get_insights(self, zone=None, tenant=None):
        """Fetches insights
            Usage: insights, status = api.get_insights()
        """
        data = {}
        if tenant is None:
            if utils.cfg["tenant"]: data["tenant.id"] = utils.cfg["tenant"]
        else:
            data["tenant.id"] = tenant
        if zone: data["zone"] = zone
        
        if g_debug: print(data)

        ret, status = self.get("api/v2/listInsights", data)
        return ret.get("insights", []), status

    def get_enforced(self, tenant=None):
        """Fetches enforced lenses
            Usage: lenses, status = api.get_enforced()
        """
        data = {}
        if tenant is None:
            if utils.cfg["tenant"]:
                data["tenant.id"] = utils.cfg["tenant"]
        else:
            data["tenant.id"] = tenant

        if g_debug: print(data)

        ret, status = self.get("api/v2/listShieldedLens", data)
        return ret.get("shielded_lens", []), status

    def set_enforced(self, op, zone=None, app=None, pod=None, container=None,
            process=None, parent=None, binpath=None, svc=None, svc_port=None, tenant=None):
        """Change the enforcement state of a lens
            Usage: status = api.set_enforced("DEL")
        """
        data = {"op": 2 if op=="DEL" else 1}
        if tenant is None:
            if utils.cfg["tenant"]:
                data["tenant"] = {"id": utils.cfg["tenant"]}
            else:
                assert False, "*** tenant not specified"
        else:
            data["tenant"] = {"id": tenant}

        if zone: data["zone"] = zone
        if app: data["app"] = app
        if svc:
            assert svc_port is not None, "both service and port need to be specified"
            data["service"] = svc
            data["service_port"] = int(svc_port)
        if container:
            assert None not in [zone, app, pod], "zone, app, and pod must be specified for the container"
            data["pod"] = pod
            data["container"] = container
        if process:
            assert None not in [zone, app, process, parent, binpath], "zone, app, process, parent, binpath must be specified"
            data["process"] = process
            data["parentProcess"] = parent
            data["binaryName"] = binpath

        if g_debug: print(data)

        ret, status = self.post("api/v2/setShieldStatus", data)
        return ret, status

    def get_fw_config(self, zone, tenant=None):
        """Fetches firewall knobs enabled for a tenant in the zone
            Usage: knobs, status = api.get_fw_config()
        """
        data = {}
        if tenant is None:
            if utils.cfg["tenant"]:
                data["tenant_id"] = utils.cfg["tenant"]
        else:
            data["tenant_id"] = tenant
        data["zone"] = zone

        if g_debug:
            print(data)

        ret, status = self.get("api/v2/getFirewallConfig", data)
        if status == 0:
            return ret.get("knobs", [])
        return None

    def update_fw_config(self, zone, tenant=None, data_file_location=None):
        """Update firewall knobs for a tenant in the zone
            Usage: status = api.update_fw_config()
        """
        if data_file_location:
            json_file_location = data_file_location
        else:
            json_file_location = "/tmp/fw_cfg.json"

        try:
            with open(json_file_location) as f:
                file_data = yaml.load(f, yaml.SafeLoader)
        except Exception as ex:
            print(ex)
            return "exception parsing json"

        if g_debug:
            print(file_data)

        data = {"knobs": file_data, "zone": zone}
        if tenant is None:
            if utils.cfg["tenant"]:
                data["tenant_id"] = utils.cfg["tenant"]
            else:
                assert False, "*** tenant not specified"
        else:
            data["tenant_id"] = tenant

        if g_debug:
            print(data)

        ret, status = self.post("api/v2/updateFirewallConfig", data)
        return status

    def get_helm_values(self, workload_name, nanny=None, tenant=None):
        """Fetches helm values for nanny/controller or firewall chart
            Usage: values_yaml = api.get_helm_values()
        """
        data = {"workload_name": workload_name}
        if tenant is None:
            if utils.cfg["tenant"]:
                data["tenant.id"] = utils.cfg["tenant"]
            else:
                assert False, "*** tenant not specified"
        else:
            data["tenant.id"] = tenant

        if nanny is None:
            data["yaml_type"] = "1"
        else:
            data["yaml_type"] = "2"

        if g_debug:
            print(data)

        ret, status = self.get("api/v2/createFortifyYaml", data)
        if g_debug:
            print(ret)

        if "code" in ret["response"] and ret["response"]["code"] != 0:
            print("error ", ret["response"]["message"])
            return ""

        return ret["workload_yaml"]

    def delete_tenant(self, tenant=None):
        """Get pod name mapping for a zone
            Usage: values_yaml = api.get_pod_mapping()
        """
        if tenant is None:
            if utils.cfg["tenant"]:
                tenant = utils.cfg["tenant"]
            else:
                assert False, "*** tenant not specified"
        else:
            tenant = tenant

        data = {
            "tenant": {
              "id": tenant,
            }
        }

        if g_debug:
            print(data)

        ret, status = self.post("api/v2/deleteTenant", data)
        return ret, status

    def get_pod_mapping(self, zone, tenant=None):
        """Get pod name mapping for a zone
            Usage: values_yaml = api.get_pod_mapping()
        """
        data = {}
        if zone: data["zone"] = zone
        if tenant is None:
            if utils.cfg["tenant"]:
                data["tenant.id"] = utils.cfg["tenant"]
            else:
                assert False, "*** tenant not specified"
        else:
            data["tenant.id"] = tenant

        if g_debug:
            print(data)

        ret, status = self.get("api/v2/listPodNameMappings", data)
        return ret.get("pod_name_mapping", []), status

    def add_pod_mapping(self, zone, app, pattern, name, tenant=None):
        """Get pod name mapping for a zone
            Usage: values_yaml = api.get_pod_mapping()
        """
        if tenant is None:
            if utils.cfg["tenant"]:
                tenant = utils.cfg["tenant"]
            else:
                assert False, "*** tenant not specified"
        else:
            tenant = tenant

        data = {
                    "tenant": {
                        "id": tenant,
                    },
                    "pod_name_mapping": [
                        {
                            "zone": zone,
                            "app": app,
                            "pod_name_regex_pattern": pattern,
                            "translated_pod_name": name
                        }
                    ]
                }

        if g_debug:
            print(data)

        ret, status = self.post("api/v2/addPodNameMappings", data)
        return ret, status

    def del_pod_mapping(self, zone, app, pattern, name, tenant=None):
        """ Delete pod name mapping entry
            Usage: values_yaml = api.get_pod_mapping()
        """
        if tenant is None:
            if utils.cfg["tenant"]:
                tenant = utils.cfg["tenant"]
            else:
                assert False, "*** tenant not specified"
        else:
            tenant = tenant

        data = {
                    "tenant": {
                        "id": tenant,
                    },
                    "pod_name_mapping": [
                        {
                            "zone": zone,
                            "app": app,
                            "pod_name_regex_pattern": pattern,
                            "translated_pod_name": name
                        }
                    ]
                }

        if g_debug:
            print(data)

        ret, status = self.post("api/v2/deletePodNameMappings", data)
        return ret, status
